import os
from pathlib import Path
from typing import List, Optional

import toml
from pydantic import BaseModel

from common.utils import get_with_home_path


class Watchmen(BaseModel):
    engine: str
    engines: List[str] = None
    log_dir: Optional[str] = None
    log_level: Optional[str] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    pid: Optional[str] = None
    mat: Optional[str] = None
    cache: Optional[str] = None


class Sock(BaseModel):
    path: str


class Socket(BaseModel):
    host: str
    port: int


class Http(BaseModel):
    host: str
    port: int


class Redis(BaseModel):
    host: str
    port: int
    username: Optional[str] = None
    password: Optional[str] = None
    queue_index: Optional[int] = None
    queue_name: Optional[str] = None
    subscribe_channels: Optional[List[str]] = None
    subscribe_name: Optional[str] = None


class Config(BaseModel):
    watchmen: Watchmen
    sock: Sock
    socket: Socket
    http: Http
    redis: Redis

    @classmethod
    def init(cls, path: Optional[str] = None) -> "Config":
        """
        Initialize config from path.
        :param path: path to config file (default: "$HOME/.watchmen/config.toml")
        :return: Config
        """
        if path is None:
            home = os.path.expanduser("~")
            path = Path(os.path.join(home, ".watchmen", "config.toml"))
        return cls.from_path(path=Path(path))

    @classmethod
    def from_path(cls, path: str) -> "Config":
        config = cls.model_validate(toml.load(path))

        # "$HOME" or "~" -> "/home/username" and create parent if not exists
        if config.watchmen.log_dir is not None:
            log_dir = get_with_home_path(path=config.watchmen.log_dir)
            parent = log_dir.parent
            if not parent.exists():
                parent.mkdir(parents=True)
            config.watchmen.log_dir = str(log_dir)

        # default log level is "info"
        if config.watchmen.log_level is not None:
            allowed = ["debug", "info", "warn", "error"]
            if config.watchmen.log_level not in allowed:
                config.watchmen.log_level = "info"

        # "$HOME" or "~" -> "/home/username" and create parent if not exists
        if config.watchmen.stdout is not None:
            stdout = get_with_home_path(path=config.watchmen.stdout)
            parent = stdout.parent
            if not parent.exists():
                parent.mkdir(parents=True)
            config.watchmen.stdout = str(stdout)

        # "$HOME" or "~" -> "/home/username" and create parent if not exists
        if config.watchmen.stderr is not None:
            stderr = get_with_home_path(path=config.watchmen.stderr)
            parent = stderr.parent
            if not parent.exists():
                parent.mkdir(parents=True)
            config.watchmen.stderr = str(stderr)

        # "$HOME" or "~" -> "/home/username" and create parent if not exists
        if config.watchmen.pid is not None:
            pid = get_with_home_path(path=config.watchmen.pid)
            parent = pid.parent
            if not parent.exists():
                parent.mkdir(parents=True)
            config.watchmen.pid = str(pid)

        # default regex pattern for task config filename is "r'^.*\.(toml|ini|json)$'"
        if config.watchmen.mat is None:
            config.watchmen.mat = r"^.*\.(toml|ini|json)$"

        # "$HOME" or "~" -> "/home/username" and create parent if not exists
        if config.sock.path is not None:
            path = get_with_home_path(path=config.sock.path)
            parent = path.parent
            if not parent.exists():
                parent.mkdir(parents=True)
            config.sock.path = str(path)

        return config



import os
import re
from argparse import Namespace
from pathlib import Path
from typing import List

import toml

from common import Config
from common.handle import Request, Response, Status
from common.task import TaskFlag
from watchmen.engine import send
from watchmen.utils import output
from watchmen.utils.file import recursive_search_files
from watchmen.utils.print_result import print_result as pr
from watchmen.utils.types import String


async def list_tasks(args: Namespace, config: Config) -> None:
    reqs: List[Request] = []

    if args.task_path is not None:
        mat: str = ""
        if args.run_regex is not None:
            mat = args.run_regex
        elif config.watchmen.mat is not None:
            mat = config.watchmen.mat
        else:
            mat = "^.*\\.(toml|ini|json)$"
        regex: re.Pattern[str] = re.compile(mat)
        matched_files: List[str] = recursive_search_files(
            args.task_path, regex)

        for configfile in matched_files:
            path = Path(configfile)
            if path.is_file():
                ext = os.path.splitext(path)[1]
                if ext == ".toml":
                    data = toml.load(path)
                    for t in data['task']:
                        reqs.append(
                            Request(command="List", data=TaskFlag.from_dict(t)))
                    pass
                elif ext == ".ini":
                    pass
                elif ext == ".json":
                    pass
                else:
                    raise Exception(
                        f'File [{path}] is not a TOML or INI or JSON file'
                    )
    elif args.task_config is not None:
        path = Path(args.task_config)
        if path.is_file():
            ext = os.path.splitext(path)[1]
            if ext == ".toml":
                data = toml.load(path)
                for t in data['task']:
                    reqs.append(
                        Request(command="List", data=TaskFlag.from_dict(t)))
                pass
            elif ext == ".ini":
                pass
            elif ext == ".json":
                pass
            else:
                raise Exception(
                    f'File [{path}] is not a TOML or INI or JSON file'
                )
    elif args.task_id is not None:
        if args.task_name is not None:
            raise Exception("Cannot use '--id' and '--name' at the same time")

        reqs.append(Request(command="List", data=TaskFlag(
            id=args.task_id,
            name=None,
            group=None,
            mat=args.task_mat,
        )))
    elif args.task_name is not None:
        reqs.append(Request(command="List", data=TaskFlag(
            id=0,
            name=args.task_name,
            group=None,
            mat=args.task_mat,
        )))
    elif args.task_group is not None:
        reqs.append(Request(command="List", data=TaskFlag(
            id=0,
            name=None,
            group=args.task_group,
            mat=args.task_mat,
        )))
    else:
        reqs.append(Request(command="List", data=None))

    if args.task_less:
        print_result_less(await send(config, reqs))
    elif args.task_more:
        print_result_more(await send(config, reqs))
    else:
        print_result(await send(config, reqs))


def print_result(res: List[Response]) -> None:
    status: List[Status] = []
    for r in res:
        if r.code != 10000:
            pr([r])
            return
        if r.data is not None and 'Status' in r.data:
            for i in r.data['Status']:
                status.append(Status(**i))

    total = 0
    total_added = 0
    total_running = 0
    total_stopped = 0
    total_waiting = 0
    total_interval = 0
    total_paused = 0

    column_id: List[str] = ["ID"]
    column_name: List[str] = ["Name"]
    column_status: List[str] = [String.white("Status")]
    column_command: List[str] = ["Command"]
    column_pid: List[str] = ["Pid"]
    column_code: List[str] = ["ExitCode"]
    column_type: List[str] = ["Type"]

    for s in status:
        total += 1
        column_id.append(s.id)
        column_name.append(s.name)
        if s.status == "added":
            total_added += 1
            column_status.append(String.magenta("added"))
        elif s.status == "running":
            total_running += 1
            column_status.append(String.green("running"))
        elif s.status == "stopped":
            total_stopped += 1
            column_status.append(String.red("stopped"))
        elif s.status == "auto restart":
            column_status.append(String.rgb("auto restart", 128, 128, 128))
        elif s.status == "waiting":
            total_waiting += 1
            column_status.append(String.blue("waiting"))
        elif s.status == "interval":
            total_interval += 1
            column_status.append(String.cyan("interval"))
        elif s.status == "paused":
            total_paused += 1
            column_status.append(String.yellow("paused"))
        elif s.status == "executing":
            column_status.append(String.green("executing"))
        else:
            column_status.append(s.status)
        cmd = s.command.split('/')
        column_command.append(cmd[-1] if len(cmd) > 0 else s.command)
        column_pid.append(s.pid or "")
        column_code.append(s.pid if s.pid is not None else "")
        column_type.append(list(s.task_type.keys())[0])

    pattern = re.compile(r'\033\[[0-9;]*m')
    max_id = max([len(pattern.sub('', str(i))) for i in column_id])
    max_name = max([len(pattern.sub('', i)) for i in column_name])
    max_status = max([len(i) for i in column_status])
    max_command = max([len(pattern.sub('', i)) for i in column_command])
    max_pid = max([len(pattern.sub('', str(i))) for i in column_pid])
    max_code = max([len(pattern.sub('', str(i))) for i in column_code])
    max_type = max([len(pattern.sub('', i)) for i in column_type])

    max_status_onlytext = max([len(pattern.sub('', i)) for i in column_status])

    max_sum = max_id + max_name + max_status_onlytext + \
        max_command + max_pid + max_code + max_type + 3 * (7 - 1) + 4

    for i in range(len(column_id)):
        output("{:-<{width}}".format("", width=max_sum))
        row = "| {: <{max_id}} | {: <{max_name}} | {: <{max_status}} | {: <{max_command}} | {: <{max_pid}} | {: <{max_code}} | {: <{max_type}} |"
        output(
            row.format(column_id[i], column_name[i], column_status[i], column_command[i], column_pid[i], column_code[i], column_type[i],
                       max_id=max_id, max_name=max_name, max_status=max_status, max_command=max_command, max_pid=max_pid, max_code=max_code, max_type=max_type)
        )
    output("{:-<{width}}".format("", width=max_sum))

    total = String.purple(total)
    total_running = String.green(total_running)
    total_stopped = String.red(total_stopped)
    total_added = String.magenta(total_added)
    total_waiting = String.blue(total_waiting)
    total_interval = String.cyan(total_interval)
    total_paused = String.yellow(total_paused)
    output(f'{total} Total: {total_running} running, {total_stopped} stopped, {total_added} added, {total_waiting} waiting, {total_interval} interval, {total_paused} paused')


def print_result_more(res: List[Response]) -> None:
    status: List[Status] = []
    for r in res:
        if r.code != 10000:
            pr([r])
            return
        if r.data is not None and 'Status' in r.data:
            for i in r.data['Status']:
                status.append(Status(**i))

    total = 0
    total_added = 0
    total_running = 0
    total_stopped = 0
    total_waiting = 0
    total_interval = 0
    total_paused = 0

    column_id: List[str] = ["ID"]
    column_group: List[str] = ["Group"]
    column_name: List[str] = ["Name"]
    column_status: List[str] = [String.white("Status")]
    column_command: List[str] = ["Command"]
    column_args: List[str] = ["Args"]
    column_pid: List[str] = ["Pid"]
    column_code: List[str] = ["ExitCode"]
    column_type: List[str] = ["Type"]

    for s in status:
        total += 1
        column_id.append(s.id)
        column_group.append(s.group or "")
        column_name.append(s.name)
        if s.status == "added":
            total_added += 1
            column_status.append(String.magenta("added"))
        elif s.status == "running":
            total_running += 1
            column_status.append(String.green("running"))
        elif s.status == "stopped":
            total_stopped += 1
            column_status.append(String.red("stopped"))
        elif s.status == "auto restart":
            column_status.append(String.rgb("auto restart", 128, 128, 128))
        elif s.status == "waiting":
            total_waiting += 1
            column_status.append(String.blue("waiting"))
        elif s.status == "interval":
            total_interval += 1
            column_status.append(String.cyan("interval"))
        elif s.status == "paused":
            total_paused += 1
            column_status.append(String.yellow("paused"))
        elif s.status == "executing":
            column_status.append(String.green("executing"))
        else:
            column_status.append(s.status)
        cmd = s.command.split('/')
        column_command.append(cmd[-1] if len(cmd) > 0 else s.command)
        column_args.append(" ".join(s.args))
        column_pid.append(s.pid or "")
        column_code.append(s.pid if s.pid is not None else "")
        column_type.append(list(s.task_type.keys())[0])

    pattern = re.compile(r'\033\[[0-9;]*m')
    max_id = max([len(pattern.sub('', str(i))) for i in column_id])
    max_group = max([len(pattern.sub('', i)) for i in column_group])
    max_name = max([len(pattern.sub('', i)) for i in column_name])
    max_status = max([len(i) for i in column_status])
    max_command = max([len(pattern.sub('', i)) for i in column_command])
    max_args = max([len(pattern.sub('', str(i))) for i in column_args])
    max_pid = max([len(pattern.sub('', str(i))) for i in column_pid])
    max_code = max([len(pattern.sub('', str(i))) for i in column_code])
    max_type = max([len(pattern.sub('', i)) for i in column_type])

    max_status_onlytext = max([len(pattern.sub('', i)) for i in column_status])

    max_sum = max_id + max_group + max_name + max_status_onlytext + \
        max_command + max_args + max_pid + \
        max_code + max_type + 3 * (9 - 1) + 4

    for i in range(len(column_id)):
        output("{:-<{width}}".format("", width=max_sum))
        row = "| {: <{max_id}} | {: <{max_group}} | {: <{max_name}} | {: <{max_status}} | {: <{max_command}} | {: <{max_args}} | {: <{max_pid}} | {: <{max_code}} | {: <{max_type}} |"
        output(
            row.format(column_id[i], column_group[i], column_name[i], column_status[i], column_command[i], column_args[i], column_pid[i], column_code[i], column_type[i],
                       max_id=max_id, max_group=max_group, max_name=max_name, max_status=max_status, max_command=max_command, max_args=max_args, max_pid=max_pid, max_code=max_code, max_type=max_type)
        )
    output("{:-<{width}}".format("", width=max_sum))

    total = String.purple(total)
    total_running = String.green(total_running)
    total_stopped = String.red(total_stopped)
    total_added = String.magenta(total_added)
    total_waiting = String.blue(total_waiting)
    total_interval = String.cyan(total_interval)
    total_paused = String.yellow(total_paused)
    output(f'{total} Total: {total_running} running, {total_stopped} stopped, {total_added} added, {total_waiting} waiting, {total_interval} interval, {total_paused} paused')


def print_result_less(res: List[Response]) -> None:
    status: List[Status] = []
    for r in res:
        if r.code != 10000:
            pr([r])
            return
        if r.data is not None and 'Status' in r.data:
            for i in r.data['Status']:
                status.append(Status(**i))

    total = 0
    total_added = 0
    total_running = 0
    total_stopped = 0
    total_waiting = 0
    total_interval = 0
    total_paused = 0

    column_id: List[str] = ["ID"]
    column_name: List[str] = ["Name"]
    column_status: List[str] = [String.white("Status")]

    for s in status:
        total += 1
        column_id.append(s.id)
        column_name.append(s.name)
        if s.status == "added":
            total_added += 1
            column_status.append(String.magenta("added"))
        elif s.status == "running":
            total_running += 1
            column_status.append(String.green("running"))
        elif s.status == "stopped":
            total_stopped += 1
            column_status.append(String.red("stopped"))
        elif s.status == "auto restart":
            column_status.append(String.rgb("auto restart", 128, 128, 128))
        elif s.status == "waiting":
            total_waiting += 1
            column_status.append(String.blue("waiting"))
        elif s.status == "interval":
            total_interval += 1
            column_status.append(String.cyan("interval"))
        elif s.status == "paused":
            total_paused += 1
            column_status.append(String.yellow("paused"))
        elif s.status == "executing":
            column_status.append(String.green("executing"))
        else:
            column_status.append(s.status)

    pattern = re.compile(r'\033\[[0-9;]*m')
    max_id = max([len(pattern.sub('', str(i))) for i in column_id])
    max_name = max([len(pattern.sub('', i)) for i in column_name])
    max_status = max([len(i) for i in column_status])

    max_status_onlytext = max([len(pattern.sub('', i)) for i in column_status])

    max_sum = max_id + max_name + max_status_onlytext + 3 * (4 - 1) + 1

    for i in range(len(column_id)):
        output("{:-<{width}}".format("", width=max_sum))
        row = "| {: <{max_id}} | {: <{max_name}} | {: <{max_status}} |"
        output(
            row.format(column_id[i], column_name[i], column_status[i],
                       max_id=max_id, max_name=max_name, max_status=max_status)
        )
    output("{:-<{width}}".format("", width=max_sum))

    total = String.purple(total)
    total_running = String.green(total_running)
    total_stopped = String.red(total_stopped)
    total_added = String.magenta(total_added)
    total_waiting = String.blue(total_waiting)
    total_interval = String.cyan(total_interval)
    total_paused = String.yellow(total_paused)
    output(f'{total} Total: {total_running} running, {total_stopped} stopped, {total_added} added, {total_waiting} waiting, {total_interval} interval, {total_paused} paused')



from typing import List

from common.config import Config
from common.handle import Request, Response
from watchmen.engine.sock import send as send_sock


async def send(config: Config, requests: List[Request]) -> List[Response]:
    if config.watchmen.engine == "sock":
        return await send_sock(config.sock.path, requests)
    else:
        raise Exception("No engine found")



from typing import List

from .arg import DaemonArgs, TaskArgs
from .config import Config
from .consts import ExitCode
from .handle import Request, Response, Status
from .log import get_logger
from .task import AsyncTask, PeriodicTask, ScheduledTask, Task, TaskFlag
from .utils import get_with_home, get_with_home_path


VERSION = "0.1.0"


__all__: List[str] = [
    "DaemonArgs", "TaskArgs",
    "Config",
    "ExitCode",
    "Request", "Response", "Status",
    "get_logger",
    "AsyncTask", "PeriodicTask", "ScheduledTask", "Task", "TaskFlag"
    "get_with_home", "get_with_home_path",
    "VERSION",
]



import os
from pathlib import Path


def get_with_home(path: str) -> str:
    home_dir: str = os.path.expanduser("~")
    if path.startswith("$HOME"):
        return path.replace("$HOME", home_dir)
    elif path.startswith("~"):
        return path.replace("~", home_dir)
    else:
        return path


def get_with_home_path(path: str) -> Path:
    return Path(get_with_home(path=path))



import asyncio
import os
import subprocess
from argparse import Namespace
from asyncio.subprocess import Process
from io import TextIOWrapper
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class ScheduledTask(BaseModel):
    year: Optional[int] = None
    month: Optional[int] = None
    day: Optional[int] = None
    hour: Optional[int] = None
    minute: Optional[int] = None
    second: Optional[int] = None


class AsyncTask(BaseModel):
    max_restart: int
    has_restart: int
    started_at: int
    stopped_at: int


class PeriodicTask(BaseModel):
    started_after: int = 0
    interval: int
    last_run: int = 0
    sync: bool = False


class Task(BaseModel):
    id: int
    name: str
    command: str
    args: List[str]
    dir: Optional[str] = None
    env: Dict[str, str]
    stdin: Optional[bool] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    created_at: int
    task_type: Any
    pid: Optional[int] = None
    status: Optional[str] = "added"
    code: Optional[int] = None

    @classmethod
    def default(cls) -> "Task":
        return cls(
            id=0,
            name="Default",
            command="",
            args=[],
            dir=None,
            env={},
            stdin=None,
            stdout=None,
            stderr=None,
            created_at=0,
            task_type=AsyncTask(max_restart=5, has_restart=0,
                                started_at=0, stopped_at=0),
            pid=None,
            status=None,
            code=None,
        )

    def into_dict(self) -> Dict[str, Any]:
        """
        Convert task to dict.
        Reconstruct data to facilitate communication with 'rust' 
        :return: Dict[str, Any]
        """
        data = self.model_dump()

        if "max_restart" in data["task_type"]:
            data["task_type"] = {"Async": data["task_type"]}
        elif "started_after" in data["task_type"]:
            data["task_type"] = {"Periodic": data["task_type"]}
        elif "year" in data["task_type"]:
            data["task_type"] = {"Scheduled": data["task_type"]}
        else:
            raise Exception("Unknown task type")
        return data

    @classmethod
    def from_args(cls, args):
        result = cls(**args)
        return result

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Task":
        """
        Convert dict to task.
        Reconstruct data to facilitate communication with 'rust'
        :param data: Dict[str, Any]
        :return: Task
        """
        result = cls(**data)
        if "Async" in result.task_type:
            result.task_type = AsyncTask(**result.task_type["Async"])
        elif "Periodic" in result.task_type:
            result.task_type = PeriodicTask(
                **result.task_type["Periodic"])
        elif "Scheduled" in result.task_type:
            result.task_type = ScheduledTask(
                **result.task_type["Scheduled"])
        else:
            raise Exception("Unknown task type")
        return result

    async def start(self) -> Process:
        stdin_file = None
        stdout_file = None
        stderr_file = None

        if self.stdin is not None:
            stdin_file: int = asyncio.subprocess.PIPE

        if self.stdout is not None:
            path = Path(self.stdout)
            parent = path.parent
            if not parent.exists():
                parent.mkdir(parents=True)
            stdout_file: TextIOWrapper = open(file=self.stdout, mode="a+")

        if self.stderr is not None:
            path = Path(self.stderr)
            parent = path.parent
            if not parent.exists():
                parent.mkdir(parents=True)
            stderr_file: TextIOWrapper = open(file=self.stderr, mode="a+")

        env: os._Environ[str] = os.environ
        if self.env is not None:
            env.update(self.env)

        child = await asyncio.create_subprocess_exec(
            *[self.command, *self.args],
            stdin=stdin_file,
            stdout=stdout_file,
            stderr=stderr_file,
            env=os.environ,
            cwd=self.dir,
            preexec_fn=subprocess.os.setsid,
        )

        return child


class TaskFlag(BaseModel):
    id: int
    name: Optional[str] = None
    group: Optional[str] = None
    mat: bool = False

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Task":
        """
        Convert dict to task.
        Reconstruct data to facilitate communication with 'rust'
        :param data: Dict[str, Any]
        :return: Task
        """
        result = cls(**data)
        return result

    @classmethod
    def from_args(cls, args: Namespace) -> "Task":
        """
        Convert args to task.
        Reconstruct data to facilitate communication with 'rust'
        :param args: Dict[str, Any]
        :return: Task
        """
        if args.task_id is not None:
            return cls(id=args.task_id, mat=args.task_mat)
        elif args.task_name is not None:
            return cls(id=0, name=args.task_name, mat=args.task_mat)
        elif args.task_group is not None:
            return cls(id=0, group=args.task_group, mat=args.task_mat)
        raise Exception("Task is none")



import asyncio
from asyncio.subprocess import Process
import atexit
import json
import os
from pathlib import Path
import re
import signal
from typing import Dict, Optional

from common.handle import Response, Status
from common.task import AsyncTask, PeriodicTask, ScheduledTask, Task, TaskFlag
from common.utils import get_with_home_path


class TaskProcess:
    def __init__(self, task: Task, joinhandle: Optional[asyncio.Task] = None, child: Optional[Process] = None) -> None:
        self.task = task
        self.joinhandle = joinhandle
        self.child = child


class Tasks:
    _instance = None

    def __init__(self) -> None:
        self._tasks: Dict[int, TaskProcess] = {}
        self._cache_path: Optional[Path] = None

        atexit.register(self._atexit)

    def _atexit(self):
        for v in self._tasks.values():
            if v.child is not None:
                try:
                    v.child.kill()
                except:
                    pass

    def set_cache_path(self, path: Path) -> None:
        self._cache_path = path

    def get_cache_path(self) -> Optional[Path]:
        return self._cache_path

    def add(self, task_id: int, tp: TaskProcess) -> None:
        self._tasks[task_id] = tp

    def check(self, task_id: int) -> bool:
        return task_id in self._tasks

    def get(self, task_id: int) -> Optional[TaskProcess]:
        return self._tasks.get(task_id)

    def get_by_name(self, name: str) -> Optional[TaskProcess]:
        for tp in self._tasks.values():
            if tp.task.name == name:
                return tp
        return None

    def get_all(self) -> Dict[int, TaskProcess]:
        return self._tasks

    def remove(self, task_id: int) -> None:
        if task_id in self._tasks:
            del self._tasks[task_id]

    def __new__(cls) -> "Tasks":
        if cls._instance is None:
            cls._instance = super(Tasks, cls).__new__(cls)
        return cls._instance


tasks = Tasks()


async def get_all() -> Dict[int, Task]:
    """
    Get all tasks.
    :return: Response
    """
    return tasks.get_all()


async def load(path: str) -> None:
    """
    Load tasks from file.
    :param path: File path
    :return: Response
    """
    path_home = get_with_home_path(path)

    tasks.set_cache_path(path_home)

    if not path_home.exists() or not path_home.is_file():
        raise Exception(f"Cache file [{path_home}] is not valid")

    with open(path_home, "r") as f:
        tasks_cache = json.load(f)

    for task in tasks_cache:
        tp = TaskProcess(task=Task.from_dict(task))
        if isinstance(tp.task.task_type, AsyncTask):
            if tp.task.status == "running":
                child = await tp.task.start()

                task_id = tp.task.id
                max_restart = tp.task.task_type.max_restart
                has_restart = tp.task.task_type.has_restart

                async def watch():
                    await child.wait()
                    print(task_id)
                    print(child.returncode)
                    print(max_restart)
                    print(has_restart)
                    returncode = child.returncode
                    if max_restart is None:
                        await update(task_id, None, "stopped", returncode, False)
                    else:
                        if max_restart == 0:
                            await update(task_id, None, "auto restart", returncode, True)
                        else:
                            if has_restart < max_restart:
                                await update(task_id, None, "auto restart", returncode, True)
                            else:
                                await update(task_id, None, "stopped", returncode, False)

                tp.joinhandle = asyncio.create_task(watch())
                tp.child = child
                tp.task.pid = child.pid
                tp.task.status = "running"
                tp.task.code = None
        tasks.add(tp.task.id, tp)


async def cache() -> None:
    path = tasks.get_cache_path()
    if path is not None:
        parent = path.parent
        if not parent.exists():
            parent.mkdir(parents=True)
        tasks_cache = []
        for tp in tasks.get_all().values():
            tasks_cache.append(tp.task.into_dict())
        with open(path, "w") as f:
            json.dump(tasks_cache, f)


async def update(task_id: int, pid: int, status: str, code: int, restart: Optional[bool] = False) -> Response:
    """
    Update task status.
    :param task_id: Task id
    :param pid: Process id
    :param status: Task status
    :param code: Exit code
    :param restart: Restart
    :return: Response
    """
    tp = tasks.get(task_id)
    if tp is None:
        return Response.failed(f"Task [{task_id}] not exists")

    tp.task.pid = pid
    tp.task.status = status
    tp.task.code = code

    if isinstance(tp.task.task_type, AsyncTask):
        has = tp.task.task_type.has_restart
        if restart is not None:
            if tp.task.task_type.max_restart is None:
                has = 0
            else:
                if has < tp.task.task_type.max_restart:
                    has += 1
        tp.task.task_type.has_restart = has

    return Response.success(f"Task [{task_id}] updated")


async def run(task: Task) -> Response:
    """
    Run task.
    :param task: Task
    :return: Response
    """
    await add(task)
    return await start(TaskFlag(id=task.id, name="", mat=False))


async def add(task: Task) -> Response:
    if tasks.check(task.id):
        return Response.failed(f"Task [{task.id}] already exists")

    tp = TaskProcess(task)
    tasks.add(task.id, tp)
    asyncio.create_task(cache())
    return Response.success(f"Task [{task.id}] added")


async def re_load(task: Task) -> Response:
    """
    Run task.
    :param task: Task
    :return: Response
    """
    await remove(TaskFlag(id=task.id, name="", mat=False), to_cache=False)
    return await add(task)


async def start(tf: TaskFlag) -> None:
    """
    Start task.
    :param tf: TaskFlag
    :return: None
    """
    tp = tasks.get(tf.id)
    if tp is None:
        raise ValueError(f"Task [{tf.id}] not exists")

    if isinstance(tp.task.task_type, AsyncTask):
        if tp.task.status == "running":
            raise ValueError(f"Task [{tf.id}] is running")

        child = await tp.task.start()

        async def watch():
            await child.wait()
            returncode = child.returncode
            if tp.task.task_type.max_restart is None:
                await update(tp.task.id, None, "stopped", returncode, restart)
            else:
                if tp.task.task_type.max_restart == 0:
                    await update(tp.task.id, None, "auto restart", returncode)
                else:
                    if tp.task.task_type.has_restart < tp.task.task_type.max_restart:
                        await update(tp.task.id, None, "auto restart", returncode)
                    else:
                        await update(tp.task.id, None, "stopped", returncode)

        tp.joinhandle = asyncio.create_task(watch())
        tp.child = child
        tp.task.pid = child.pid
        tp.task.status = "running"
        tp.task.code = None

        asyncio.create_task(cache())

        return Response.success(f"Task [{tf.id}] started")
    elif isinstance(tp.task.task_type, PeriodicTask):
        print("This task type is currently not supported")
    elif isinstance(tp.task.task_type, ScheduledTask):
        print("This task type is currently not supported")
    raise ValueError("Task type not supported")


async def stop(tf: TaskFlag, to_cache: bool = True) -> Response:
    tp = tasks.get(tf.id)
    if tp is None:
        raise ValueError(f"Task [{tf.id}] not exists")

    if tp.task.status != "running" and tp.task.status != "auto restart":
        raise ValueError(f"Task [{tf.id}] is not running")

    pid = tp.task.pid
    if pid is None:
        raise ValueError(f"Task [{tf.id}] is not running")

    try:
        os.kill(pid, signal.SIGTERM)
        tp.task.status = "stopped"
        if to_cache:
            asyncio.create_task(cache())
        return Response.success(f"Task [{tf.id}] stopped")
    except ProcessLookupError:
        raise ValueError(f"Task [{tf.id}] is not running")


async def restart(tf: TaskFlag) -> Response:
    await stop(tf, False)
    return await start(tf)


async def remove(tf: TaskFlag, to_cache: bool = True) -> Response:
    if tf.id > 0:
        tp = tasks.get(tf.id)
    else:
        tp = tasks.get_by_name(tf.name)

    if tp is None:
        raise ValueError(f"Task [{tf.id}] not exists")

    if tp.task.status == "running":
        raise ValueError("Task is running, please stop it first")

    tasks.remove(tp.task.id)

    if to_cache:
        asyncio.create_task(cache())

    return Response.success(f"Task [{tf.id}] removed")


async def lst(condition: Optional[TaskFlag]) -> Response:
    """
    List task.
    :param condition: Optional[TaskFlag]
    :return: Response
    """
    if condition is None:
        # list all task status
        statuses = []
        for k, v in tasks.get_all().items():
            status = Status(
                id=k,
                name=v.task.name,
                command=v.task.command,
                args=v.task.args,
                dir=v.task.dir,
                env=v.task.env,
                stdin=v.task.stdin,
                stdout=v.task.stdout,
                stderr=v.task.stderr,
                created_at=v.task.created_at,
                task_type=v.task.task_type,
                pid=v.task.pid,
                status=v.task.status,
                exit_code=v.task.code,
            )
            statuses.append(status)
        return Response.success(statuses)
    else:
        # list task status by condition
        if condition.id > 0:
            # condition by id
            tp = tasks.get(condition.id)
            if tp is None:
                return Response.success([])
            status = Status(
                id=tp.task.id,
                name=tp.task.name,
                command=tp.task.command,
                args=tp.task.args,
                dir=tp.task.dir,
                env=tp.task.env,
                stdin=tp.task.stdin,
                stdout=tp.task.stdout,
                stderr=tp.task.stderr,
                created_at=tp.task.created_at,
                task_type=tp.task.task_type,
                pid=tp.task.pid,
                status=tp.task.status,
                exit_code=tp.task.code,
            )
            return Response.success([status])
        elif condition.mat:
            # condition by name with regex
            statuses = []
            for k, v in tasks.get_all().items():
                if re.match(condition.name, v.task.name):
                    status = Status(
                        id=k,
                        name=v.task.name,
                        command=v.task.command,
                        args=v.task.args,
                        dir=v.task.dir,
                        env=v.task.env,
                        stdin=v.task.stdin,
                        stdout=v.task.stdout,
                        stderr=v.task.stderr,
                        created_at=v.task.created_at,
                        task_type=v.task.task_type,
                        pid=v.task.pid,
                        status=v.task.status,
                        exit_code=v.task.code,
                    )
                    statuses.append(status)
            return Response.success(statuses)
        else:
            # condition by name
            tp = tasks.get_by_name(condition.name)
            if tp is None:
                return Response.success([])
            status = Status(
                id=tp.task.id,
                name=tp.task.name,
                command=tp.task.command,
                args=tp.task.args,
                dir=tp.task.dir,
                env=tp.task.env,
                stdin=tp.task.stdin,
                stdout=tp.task.stdout,
                stderr=tp.task.stderr,
                created_at=tp.task.created_at,
                task_type=tp.task.task_type,
                pid=tp.task.pid,
                status=tp.task.status,
                exit_code=tp.task.code,
            )
            return Response.success([status])



import json
from common.handle import Request


class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Request):
            return obj.into_dict()
        else:
            return json.JSONEncoder.default(obj)



from argparse import Namespace
from typing import List

from common import Config
from common.handle import Request, Response
from common.task import Task
from watchmen.commands.base import task_to_request
from watchmen.engine import send
from watchmen.utils.print_result import print_result


async def reload_task(args: Namespace, config: Config) -> None:
    tasks: List[Task] = await task_to_request(args, config)
    if len(tasks) == 0:
        print_result(Response.wrong("No task to reload"))
    else:
        requests: List[Request] = []
        for t in tasks:
            requests.append(Request(command="Reload", data=t))
        print_result(await send(config, requests))



class String(str):
    def rgb(self, r, g, b) -> str:
        return f"\033[38;2;{r};{g};{b}m{self}\033[0m"

    def red(self) -> str:
        return f"\033[31m{self}\033[0m"

    def green(self) -> str:
        return f"\033[32m{self}\033[0m"

    def yellow(self) -> str:
        return f"\033[33m{self}\033[0m"

    def blue(self) -> str:
        return f"\033[34m{self}\033[0m"

    def purple(self) -> str:
        return f"\033[35m{self}\033[0m"

    def cyan(self) -> str:
        return f"\033[36m{self}\033[0m"

    def white(self) -> str:
        return f"\033[37m{self}\033[0m"

    def black(self) -> str:
        return f"\033[38m{self}\033[0m"

    def magenta(self) -> str:
        return f"\033[95m{self}\033[0m"

    def bold(self) -> str:
        return f"\033[1m{self}\033[0m"

    def italic(self) -> str:
        return f"\033[3m{self}\033[0m"

    def underline(self) -> str:
        return f"\033[4m{self}\033[0m"

    def blink(self) -> str:
        return f"\033[5m{self}\033[0m"

    def reverse(self) -> str:
        return f"\033[7m{self}\033[0m"

    def invisible(self) -> str:
        return f"\033[8m{self}\033[0m"

    def strikethrough(self) -> str:
        return f"\033[9m{self}\033[0m"



import asyncio

from common import Config, get_logger
from watchmend.lib import load


async def start(config: Config, load_cache: bool) -> None:
    logger = get_logger(log_dir=config.watchmen.log_dir)
    if load_cache:
        try:
            await load(path=config.watchmen.cache)
        except Exception as e:
            logger.warning(f"Failed to load cache: {e}")

    task_sock = None
    if "sock" in config.watchmen.engines:
        from .sock import start as start_sock
        logger.info("Starting sock...")
        task_sock = await start_sock(config=config)

    tasks = []
    if task_sock is not None:
        tasks.append(task_sock)

    await asyncio.gather(*tasks)



import os
import re
from argparse import Namespace
from pathlib import Path
from typing import List

import toml

from common import Config
from common.task import Task, TaskFlag
from watchmen.utils.file import recursive_search_files


async def task_to_request(args: Namespace, config: Config) -> List[Task]:
    tasks: List[Task] = []

    if args.task_path is not None:
        mat: str = ""
        if args.run_regex is not None:
            mat = args.run_regex
        elif config.watchmen.mat is not None:
            mat = config.watchmen.mat
        else:
            mat = "^.*\\.(toml|ini|json)$"
        regex: re.Pattern[str] = re.compile(mat)
        matched_files: List[str] = recursive_search_files(
            args.task_path, regex)

        for configfile in matched_files:
            path = Path(configfile)
            if path.is_file():
                ext = os.path.splitext(path)[1]
                if ext == ".toml":
                    data = toml.load(path)
                    for t in data['task']:
                        tasks.append(Task.from_dict(t))
                    pass
                elif ext == ".ini":
                    pass
                elif ext == ".json":
                    pass
                else:
                    raise Exception(
                        f'File [{path}] is not a TOML or INI or JSON file'
                    )
        return tasks
    elif args.task_config is not None:
        path = Path(args.task_config)
        if path.is_file():
            ext = os.path.splitext(path)[1]
            if ext == ".toml":
                data = toml.load(path)
                for t in data['task']:
                    tasks.append(Task.from_dict(t))
                pass
            elif ext == ".ini":
                pass
            elif ext == ".json":
                pass
            else:
                raise Exception(
                    f'File [{path}] is not a TOML or INI or JSON file'
                )
        return tasks
    else:
        return []


async def taskflag_to_request(args: Namespace, config: Config) -> List[TaskFlag]:
    tfs: List[TaskFlag] = []

    if args.task_path is not None:
        mat: str = ""
        if args.run_regex is not None:
            mat = args.run_regex
        elif config.watchmen.mat is not None:
            mat = config.watchmen.mat
        else:
            mat = "^.*\\.(toml|ini|json)$"
        regex: re.Pattern[str] = re.compile(mat)
        matched_files: List[str] = recursive_search_files(
            args.task_path, regex)

        for configfile in matched_files:
            path = Path(configfile)
            if path.is_file():
                ext = os.path.splitext(path)[1]
                if ext == ".toml":
                    data = toml.load(path)
                    for t in data['task']:
                        tfs.append(TaskFlag.from_dict(t))
                    pass
                elif ext == ".ini":
                    pass
                elif ext == ".json":
                    pass
                else:
                    raise Exception(
                        f'File [{path}] is not a TOML or INI or JSON file'
                    )
    elif args.task_config is not None:
        path = Path(args.task_config)
        if path.is_file():
            ext = os.path.splitext(path)[1]
            if ext == ".toml":
                data = toml.load(path)
                for t in data['task']:
                    tfs.append(TaskFlag.from_dict(t))
                pass
            elif ext == ".ini":
                pass
            elif ext == ".json":
                pass
            else:
                raise Exception(
                    f'File [{path}] is not a TOML or INI or JSON file'
                )
    else:
        tfs.append(TaskFlag.from_args(args))
    return tfs



import os
from pathlib import Path

from common import ExitCode
from watchmen.utils import output


CONFIG = """[watchmen]
# The engine to use for the watchmen server
# Valid values are "sock", "socket", "http", "redis"
# sock: Unix socket
# socket: TCP socket
# http: HTTP Api (Include Web panel)
# redis: Redis pub/sub
engines = ["sock"]

# The default engine to use for connecting to the watchmen server
engine = "sock"

# The log directory of the watchmen server
log_dir = "$HOME/.watchmen/logs"

# The log level of the watchmen server
# Valid values are "debug", "info", "warn", "error". Default is "info"
log_level = "info"

# The standard output of the watchmen server
# Default is None
stdout = "$HOME/.watchmen/watchmen.stdout.log"

# The standard error of the watchmen server
# Default is None
stderr = "$HOME/.watchmen/watchmen.stderr.log"

# The pid file of the watchmen server
# Default is `$HOME/.watchmen/watchmen.pid`
pid = "$HOME/.watchmen/watchmen.pid"

# The task config file name matching pattern
# Default is `^.*\\.(toml|ini|json)$`
mat = "^.*\\.(toml|ini|json)$"

# Tasks cache file, json format
cache = "$HOME/.watchmen/cache.json"

# Monitor interval for rerun tasks, u64: second
interval = 5


[sock]
# The unix socket path of the watchmen server
path = "/tmp/watchmen.sock"


[socket]
host = "127.0.0.1"
port = 1949


[http]
host = "127.0.0.1"
port = 1997


[redis]
host = "localhost"
port = 6379
username = ""
password = ""
queue_index = 0
queue_name = "watchmen"
subscribe_channels = ["watchmen"]
subscribe_name = "watchmen"
"""


def generate(path: str) -> int:
    config_path = None
    if path == "":
        home = os.path.expanduser("~")
        config_path = Path(os.path.join(home, ".watchmen", "config.toml"))
    else:
        config_path = Path(os.getcwd()).joinpath(path)
        if config_path.is_dir():
            config_path = config_path.joinpath("config.toml")
        else:
            ext = config_path.suffix
            if ext != ".toml":
                raise ValueError("Config file must be toml")

    parent = config_path.parent
    if not parent.exists():
        parent.mkdir(parents=True)

    output(f"Generate config file to {config_path}")

    with open(config_path, "w") as f:
        f.write(CONFIG)

    return ExitCode.SUCCESS



import asyncio
from argparse import Namespace

from common import Config, TaskArgs, ExitCode, VERSION
from watchmen.args import generate
from watchmen.commands import handle_exec
from watchmen.utils import output
from watchmen.utils.types import String


async def _main() -> int:
    clargs: Namespace = TaskArgs.parse()
    if clargs.version:
        output(String(f"Watchmen python {VERSION}").green())
        return ExitCode.SUCCESS

    if clargs.generate is not None:
        try:
            return generate(path=clargs.generate)
        except ValueError as e:
            output(String(e).red())
            return ExitCode.ERROR

    config: Config = Config.init(path=clargs.config)

    if clargs.subcommand is not None:
        if clargs.engine is not None:
            config.watchmen.engine = clargs.engine

        await handle_exec(clargs, config)

        # try:
        #     await handle_exec(clargs, config)
        # except Exception as e:
        #     output(String(e).red())
        #     return ExitCode.ERROR


def main() -> int:
    asyncio.run(_main())



import asyncio

from .engine import start
from .monitor import run_monitor
from common import Config, DaemonArgs, ExitCode, VERSION


async def _main(config: Config, load: bool) -> int:
    asyncio.create_task(run_monitor())
    await start(config=config, load_cache=load)


def main() -> int:
    clargs = DaemonArgs.parse()
    if clargs.version:
        print(f"\033[32mWatchmen python {VERSION}\033[0m")
        return ExitCode.SUCCESS

    load: bool = clargs.load

    config: Config = Config.init(path=clargs.config)

    try:
        asyncio.run(_main(config=config, load=load))
    except KeyboardInterrupt:
        return ExitCode.SUCCESS




import os
import re
from typing import List


def recursive_search_files(foldr: str, regex: re.Pattern[str]) -> List[str]:
    matched_files: List[str] = []
    for root, _dirs, files in os.walk(foldr):
        for file in files:
            if re.match(regex, file):
                matched_files.append(os.path.join(root, file))
    return matched_files



from argparse import _SubParsersAction, ArgumentParser, Namespace


DESCRIPTION = """Watchmen (Python) is a daemon process manager that for you manage and keep your application online 24/7"""
EPILOG = """See "watchmen COMMAND --help" for more information on a specific command."""


class TaskArgs:

    def __init__(self) -> None:
        self._parser: ArgumentParser = self._create_parser()

        self._subparser = self._create_subparser()

        self._create_task_command("run", "Add and run tasks")
        self._create_task_command("add", "Add tasks")
        self._create_task_command("reload", "Reload tasks")

        self._create_flag_command("start", "Start tasks")
        self._create_flag_command("restart", "Restart tasks")
        self._create_flag_command("stop", "Stop tasks")
        self._create_flag_command("remove", "Remove tasks")

        self._create_flag_command("pause", "Pause tasks")
        self._create_flag_command("resume", "Resume tasks")

        self._create_list_command("list", "Get tasks list")

    def _create_parser(self) -> ArgumentParser:
        parser = ArgumentParser(description=DESCRIPTION,
                                usage="watchmen [OPTIONS] [COMMAND]", epilog=EPILOG)
        parser.add_argument("-c", "--config", type=str, metavar="<CONFIG>",
                                  default=None, help="Config file path. Default: $HOME/.watchmen/config.toml")
        parser.add_argument("-g", "--generate", type=str, metavar="<GENERATE>",
                                  default=None, help="Generate config file")
        parser.add_argument("-e", "--engine", type=str, metavar="<ENGINE>",
                                  default=None, help="Engine for send message", choices=["sock", "socket", "http", "redis"])
        parser.add_argument("-v", "--version", action="store_true",
                                  default=False, help="Print version")
        return parser

    def _create_subparser(self):
        subparser: _SubParsersAction[ArgumentParser] = self._parser.add_subparsers(
            title="Sub Commands", dest="subcommand")
        return subparser

    def _create_task_command(self, name: str, help_text: str) -> None:
        parser: ArgumentParser = self._subparser.add_parser(name=name, help=help_text,
                                                            usage=f"watchmen {name} [OPTIONS]")
        parser.add_argument("-p", "--path", type=str, metavar="<PATH>",
                            default=None, help="Task config directory", dest=f"task_path")
        parser.add_argument("-r", "--regex", type=str, metavar="<REGEX>",
                            default=r"^.*\.(toml|ini|json)$", help="Task config filename regex pattern", dest=f"task_regex")
        parser.add_argument("-f", "--config", type=str, metavar="<CONFIG>",
                            default=None, help="Task config file", dest=f"task_config")
        parser.add_argument("-n", "--name", type=str, metavar="<NAME>",
                            default=None, help="Task name (unique)", dest=f"task_name")
        parser.add_argument("-c", "--command", type=str, metavar="<COMMAND>",
                            default=None, help="Task command", dest=f"task_command")
        parser.add_argument("-a", "--args", nargs="+", metavar="<ARGS>",
                            default=None, help="Task arguments", dest=f"task_args")
        parser.add_argument("-d", "--dir", type=str, metavar="<DIR>",
                            default=None, help="Task working directory", dest=f"task_dir")
        parser.add_argument("-e", "--env", nargs="+", metavar="<ENV>",
                            default=None, help="Task environment variables", dest=f"task_env")
        parser.add_argument("-i", "--stdin", action="store_true",
                            default=False, help="Task standard input", dest=f"task_stdin")
        parser.add_argument("-o", "--stdout", type=str, metavar="<STDOUT>",
                            default=None, help="Task standard output", dest=f"task_stdout")
        parser.add_argument("-w", "--stderr", type=str, metavar="<STDERR>",
                            default=None, help="Task standard error", dest=f"task_stderr")

    def _create_flag_command(self, name: str, help_text: str) -> None:
        parser: ArgumentParser = self._subparser.add_parser(name=name, help=help_text,
                                                            usage=f"watchmen {name} [OPTIONS]")
        parser.add_argument("-p", "--path", type=str, metavar="<PATH>",
                            default=None, help="Task config directory", dest=f"task_path")
        parser.add_argument("-r", "--regex", type=str, metavar="<REGEX>",
                            default=r"^.*\.(toml|ini|json)$", help="Task config filename regex pattern", dest=f"task_regex")
        parser.add_argument("-f", "--config", type=str, metavar="<CONFIG>",
                            default=None, help="Task config file", dest=f"task_config")
        parser.add_argument("-i", "--id", type=int, metavar="<ID>",
                            default=None, help="Task id (unique)", dest=f"task_id")
        parser.add_argument("-n", "--name", type=str, metavar="<NAME>",
                            default=None, help="Task name (unique)", dest=f"task_name")
        parser.add_argument("-m", "--mat", action="store_true",
                            default=False, help="Is match regex pattern by namae", dest=f"task_mat")

    def _create_list_command(self, name: str, help_text: str) -> None:
        parser: ArgumentParser = self._subparser.add_parser(name=name, help=help_text,
                                                            usage=f"watchmen {name} [OPTIONS]")
        parser.add_argument("-p", "--path", type=str, metavar="<PATH>",
                            default=None, help="Task config directory", dest=f"task_path")
        parser.add_argument("-r", "--regex", type=str, metavar="<REGEX>",
                            default=r"^.*\.(toml|ini|json)$", help="Task config filename regex pattern", dest=f"task_regex")
        parser.add_argument("-f", "--config", type=str, metavar="<CONFIG>",
                            default=None, help="Task config file", dest=f"task_config")
        parser.add_argument("-i", "--id", type=int, metavar="<ID>",
                            default=None, help="Task id (unique)", dest=f"task_id")
        parser.add_argument("-n", "--name", type=str, metavar="<NAME>",
                            default=None, help="Task name (unique)", dest=f"task_name")
        parser.add_argument("-g", "--group", type=str, metavar="<GROUP>",
                            default=None, help="Task group", dest=f"task_group")
        parser.add_argument("-R", "--mat", action="store_true",
                            default=False, help="Is match regex pattern by name", dest=f"task_mat")
        parser.add_argument("-m", "--more", action="store_true",
                            default=False, help="Show more info", dest=f"task_more")
        parser.add_argument("-l", "--less", action="store_true",
                            default=False, help="Show less info", dest=f"task_less")

    @classmethod
    def parse(cls) -> Namespace:
        this = cls()
        return this._parser.parse_args()


class DaemonArgs:

    def __init__(self) -> None:
        self._parser: ArgumentParser = self._create_parser()

    def _create_parser(self) -> ArgumentParser:
        parser = ArgumentParser(description=DESCRIPTION,
                                usage="watchmen [OPTIONS]")
        parser.add_argument("-c", "--config", type=str, metavar="<CONFIG>",
                                  default=None, help="Config file path. Default: $HOME/.watchmen/config.toml")
        parser.add_argument("-l", "--load", action="store_true",
                                  default=True, help="Load cached tasks")
        parser.add_argument("-v", "--version", action="store_true",
                                  default=False, help="Print version")
        return parser

    @classmethod
    def parse(cls) -> Namespace:
        this = cls()
        return this._parser.parse_args()



import asyncio
import json
from pathlib import Path

from common.config import Config
from common.handle import Request, Response, Status
from common.task import AsyncTask
from watchmend.command import handle_exec


async def start(config: Config) -> asyncio.Task[None]:
    return asyncio.create_task(run_sock(path=config.sock.path))


async def run_sock(path: str) -> None:
    sock_path = Path(path)

    if sock_path.exists() and sock_path.is_socket():
        sock_path.unlink()

    server: asyncio.Server = await asyncio.start_unix_server(handle_connection, path=path)
    async with server:
        await server.serve_forever()


async def handle_connection(reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
    buf = await reader.read(102400)

    text = buf.decode()
    responses = []
    for i in json.loads(text):
        response = await handle_exec(Request.from_dict(i))
        responses.append(response)

    b = json.dumps([i.into_dict() for i in responses]).encode("utf-8")

    writer.write(b)
    await writer.drain()

    writer.close()
    await writer.wait_closed()



from typing import List

from common.handle import Response
from watchmen.utils import output
from watchmen.utils.types import String


def print_result(res: List[Response]) -> None:
    for r in res:
        data = r.data
        if 'String' in r.data:
            data = r.data['String']
        result = f'{r.code}\t{r.msg}\t{data}'
        if r.code == 10000:
            output(String.green(result))
        elif r.code == 40000:
            output(String.yellow(result))
        elif r.code == 50000:
            output(String.red(result))
        else:
            output(result)



from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel

from common.task import AsyncTask, PeriodicTask, ScheduledTask, Task, TaskFlag


class Request(BaseModel):
    command: str
    data: Union[Task, TaskFlag, Optional[TaskFlag],
                Tuple[TaskFlag, str]] = None

    def into_dict(self) -> Dict[str, Any]:
        """
        Convert request to dict.
        Reconstruct data to facilitate communication with 'rust'
        :return: Dict[str, Any]
        """
        data = self.model_dump()
        if data['data'] is None:
            return {"command": {"List": None}}
        if "task_type" not in data['data']:
            result = {"command": {}}
            result["command"][data["command"]] = data["data"]
            return result
        if isinstance(data["data"], dict):
            if "max_restart" in data["data"]["task_type"]:
                data["data"]["task_type"] = {
                    "Async": data["data"]["task_type"]}
            elif "interval" in data["data"]["task_type"]:
                data["data"]["task_type"] = {
                    "Periodic": data["data"]["task_type"]}
            elif "year" in data["data"]["task_type"]:
                data["data"]["task_type"] = {
                    "Scheduled": data["data"]["task_type"]}
            else:
                raise Exception("Unknown task type")
        result = {"command": {}}
        result["command"][data["command"]] = data["data"]
        return result

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Request":
        """
        Convert dict to request.
        Reconstruct data to facilitate communication with 'rust'
        :param data: Dict[str, Any]
        :return: Request
        """
        command = list(data["command"].keys())[0]
        data = data["command"][command]
        if command in ["Run", "Add", "Reload"]:
            if 'Async' in data["task_type"]:
                data["task_type"] = AsyncTask(
                    **data["task_type"]["Async"])
            elif 'Periodic' in data["task_type"]:
                data["task_type"] = PeriodicTask(
                    **data["task_type"]["Periodic"])
            elif 'Scheduled' in data["task_type"]:
                data["task_type"] = ScheduledTask(
                    **data["task_type"]["Scheduled"])
            else:
                raise Exception("Unknown task type")
            return cls(command=command, data=data)
        elif command in ["List"]:
            if data is None:
                return cls(command=command, data=None)
            else:
                return cls(command=command, data=TaskFlag(**data))
        else:
            return cls(command=command, data=TaskFlag(**data))


class Status(BaseModel):
    id: int
    group: Optional[str] = ""
    name: str
    command: str
    args: List[str]
    dir: Optional[str] = None
    env: Dict[str, str]
    stdin: Optional[bool] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    created_at: int
    task_type: Any
    pid: Optional[int] = None
    status: Optional[str] = None
    exit_code: Optional[int] = None


class Response(BaseModel):
    code: int
    msg: str
    data: Any

    def __init__(self, code: int, msg: str, data: Any = None) -> None:
        super().__init__(code=code, msg=msg, data=data)

    @classmethod
    def success(cls, data: Any = None) -> "Response":
        return cls(code=10000, msg="Success", data=data)

    @classmethod
    def wrong(cls, msg: str) -> "Response":
        return cls(code=40000, msg="Wrong", data=msg)

    @classmethod
    def failed(cls, msg: str) -> "Response":
        return cls(code=50000, msg="Failed", data=msg)

    def into_dict(self) -> Dict[str, Any]:
        """
        Convert response to dict.
        Reconstruct data to facilitate communication with 'rust' 
        :return: Dict[str, Any]
        """
        if isinstance(self.data, list):
            for item in self.data:
                if isinstance(item.task_type, AsyncTask):
                    item.task_type = {"Async": item.task_type}
                elif isinstance(item.task_type, PeriodicTask):
                    item.task_type = {"Periodic": item.task_type}
                elif isinstance(item.task_type, ScheduledTask):
                    item.task_type = {"Scheduled": item.task_type}
                else:
                    raise Exception("Unknown task type")
            self.data = {"Status": self.data}
        elif isinstance(self.data, str):
            self.data = {"String": self.data}
        return self.model_dump()

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Response":
        """
        Convert dict to response.
        Reconstruct data to facilitate communication with 'rust'
        :param data: Dict[str, Any]
        :return: Response
        """
        result = cls(**data)
        if isinstance(result.data, dict) and "Status" in result.data:
            for item in result.data["Status"]:
                if "Async" in item["task_type"]:
                    item["task_type"] = AsyncTask(
                        **item["task_type"]["Async"])
                elif "Periodic" in item["task_type"]:
                    item["task_type"] = PeriodicTask(
                        **item["task_type"]["Periodic"])
                elif "Scheduled" in item["task_type"]:
                    item["task_type"] = ScheduledTask(
                        **item["task_type"]["Scheduled"])
                else:
                    raise Exception("Unknown task type")
            result.data = [Status(**i) for i in result.data["Status"]]
        elif isinstance(result.data, dict) and "String" in result.data:
            result.data = result.data["String"]
        return result




import asyncio
import time

from .lib import get_all, start
from common.task import AsyncTask, TaskFlag


async def rerun_tasks() -> None:
    tasks = await get_all()
    for _id, tp in tasks.items():
        if isinstance(tp.task.task_type, AsyncTask):
            if tp.task.status == 'auto restart':
                await start(TaskFlag(id=_id, name=tp.task.name, mat=False))


async def run_monitor() -> None:
    interval = 5
    while True:
        start_time = time.time()
        await rerun_tasks()
        elapsed_time = time.time() - start_time
        wait_time = max(0, interval - elapsed_time)
        await asyncio.sleep(wait_time)



from argparse import Namespace

from common import Config
from watchmen.commands.run import run
from watchmen.commands.add import add
from watchmen.commands.reload import reload_task
from watchmen.commands.start import start
from watchmen.commands.restart import restart
from watchmen.commands.stop import stop
from watchmen.commands.remove import remove
from watchmen.commands.pause import pause
from watchmen.commands.resume import resume
from watchmen.commands.list import list_tasks


async def handle_exec(commands: Namespace, config: Config) -> None:
    match commands.subcommand:
        case 'run':
            return await run(commands, config)
        case 'add':
            return await add(commands, config)
        case 'reload':
            return await reload_task(commands, config)
        case 'start':
            return await start(commands, config)
        case 'restart':
            return await restart(commands, config)
        case 'stop':
            return await stop(commands, config)
        case 'remove':
            return await remove(commands, config)
        case 'pause':
            return await pause(commands, config)
        case 'resume':
            return await resume(commands, config)
        case 'list':
            return await list_tasks(commands, config)
        case _:
            raise NotImplementedError(
                f"Command {commands.subcommand} not implemented"
            )



import logging
import logging.handlers
from pathlib import Path


class Logger:
    __instance = None

    def __new__(cls, path: Path) -> "Logger":
        if not Logger.__instance:
            if not path.exists():
                path.mkdir(parents=True, exist_ok=True)

            path = path.joinpath("watchmen.log")

            Logger.__instance = object.__new__(cls)
            Logger.__instance.logger = logging.getLogger('watchmen')
            Logger.__instance.logger.setLevel(logging.DEBUG)

            console_handler = logging.StreamHandler()
            console_handler.setLevel(logging.DEBUG)

            time_rotating_handler = logging.handlers.TimedRotatingFileHandler(
                filename=path, when='midnight', interval=1, backupCount=7, encoding='utf-8')
            time_rotating_handler.setLevel(logging.INFO)

            formatter = logging.Formatter(
                '%(asctime)s - %(levelname)s - %(message)s')

            console_handler.setFormatter(formatter)
            time_rotating_handler.setFormatter(formatter)

            Logger.__instance.logger.addHandler(console_handler)
            Logger.__instance.logger.addHandler(time_rotating_handler)
        return Logger.__instance

    def debug(self, *args, **kwargs):
        self.logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        self.logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self.logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        self.logger.error(*args, **kwargs)

    def critical(self, *args, **kwargs):
        self.logger.critical(*args, **kwargs)

    def exception(self, *args, **kwargs):
        self.logger.exception(*args, **kwargs)


def get_logger(log_dir: str) -> Logger:
    return Logger(path=Path(log_dir))



import json
import socket
from pathlib import Path
from typing import List

from common.handle import Request, Response
from watchmen.utils.serialize import CustomEncoder


async def send(path: str, requests: List[Request]) -> List[Response]:
    sock_path = Path(path)
    if not sock_path.exists():
        raise Exception(f'Socket file {path} not exists')

    client_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    client_socket.connect(path)

    client_message = json.dumps(requests, cls=CustomEncoder)
    client_socket.send(client_message.encode('utf-8'))

    server_response = client_socket.recv(10240).decode('utf-8')

    response: List[Response] = []
    for i in json.loads(server_response):
        response.append(Response(i['code'], i['msg'], i['data']))

    client_socket.close()

    return response



from setuptools import setup
from setuptools import find_packages
from watchmen import VERSION

with open("README_EN.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="watchmenpy",
    version=VERSION,
    author="ahriknow",
    author_email="ahriknow@ahriknow.com",
    description="Watchmen is a daemon process manager that for you manage and keep your application online 24/7",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    keywords=[],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "watchmen=watchmen.main:main",
            "watchmend=watchmend.main:main",
        ],
    },
    install_requires=[
        "toml~=0.10.0",
        "pydantic~=2.1.0"
    ],
    extras_require={
        "redis": ["redis >= 4.5.0, < 5.0"],
    },
    python_requires=">=3.5",
)



from common import Request, Response
from watchmend.lib import run, add, re_load, start, stop, restart, remove, lst


async def handle_exec(request: Request) -> Response:
    try:
        if request.command == "Run":
            return await run(request.data)
        elif request.command == "Add":
            return await add(request.data)
        elif request.command == "Reload":
            return await re_load(request.data)
        elif request.command == "Start":
            return await start(request.data)
        elif request.command == "Stop":
            return await stop(request.data)
        elif request.command == "Restart":
            return await restart(request.data)
        elif request.command == "Remove":
            return await remove(request.data)
        elif request.command == "List":
            return await lst(request.data)
    except ValueError as e:
        return Response.failed(str(e))
